@@ -1,14 +1,8 @@
-
-'''
-The params module handles COMETS simulation parameters.
-For more information see https://segrelab.github.io/comets-manual/
-'''
-
 import pandas as pd
 import os
 
 
-def isfloat(value):
+def __isfloat(value):
     try:
         float(value)
         return True
@@ -18,26 +12,118 @@ def isfloat(value):
 
 class params:
     '''
-    Class storing COMETS parameters
+    object storing simulation-related and some default biological parameters
+    
+    The params object is an essential object needed to start a comets
+    simulation, along with a layout object containing models. The params 
+    object stores simulation information such as timeStep and maxCycles, as 
+    well as many default biological parameters such as defaultKm and 
+    defaultDiffC (the default metabolite diffusion constant). 
+    
+    All of the parameters are stored in a dictionary called 'all_params' and
+    can either be adjusted directly or using set_param(). 
+    
+    When params are written to a COMETS object for COMETS to process, two
+    files are generated ('.current_global' and 'current_package')
+    
+    Parameters
+    ----------
+    global_params : str, optional
+        optional path to an existing 'global_params' file to load
+    package_params : str, optional
+        optional path to an existing 'package_params' file to load
+        
+    Examples
+    --------
+    
+    >>> # make a params object and change two params
+    >>> import cometspy as c
+    >>> params = c.params()
+    >>> params.set_param("timeStep", 0.01) # hours
+    >>> params.set_param("spaceWidth", 0.5) # cm
+    >>> # view all the params
+    >>> params.show_params()
+    >>> # access the params directly
+    >>> params.all_params # this is a dict
+    >>> # use a params object in a comets simulation
+    >>> import cobra.test
+    >>> model = c.model(cobra.test.create_test_model("textbook"))
+    >>> model.initial_pop = [0, 0, 1.e-7]
+    >>> model.open_exchanges()
+    >>> layout = c.layout([model])
+    >>> # normally we'd now alter the media in the layout
+    >>> layout.add_typical_trace_metabolites()
+    >>> layout.set_specific_metabolite('lcts_e', 0.05) # mmol
+    >>> sim = c.comets(layout, params)
+    >>> sim.run()
+    
+    Attributes
+    ----------
+    
+    all_params : dict
+        contains every possible param as keys, with their value as the value
+    params_units : dict
+        dictionary containing the units for each possible parameter
+
     '''
     def show_params(self):
+        """
+        utility function to show all possible parameters and their units
+        """
+        
         pdparams = pd.DataFrame({'VALUE': pd.Series(self.all_params),
                                  'UNITS': pd.Series(self.param_units)})
         return pdparams
 
-    def set_param(self, name, value):
+    def set_param(self, name : str, value):
+        """
+        alters a specific parameter value
+        
+        See show_params() for a list of possible parameters to adjust. 
+        
+        Parameters
+        ----------
+        name : str
+            the name of a specific parameter to change 
+        value : variable
+            the type of the value depends on the parameter.
+
+        Examples
+        --------
+        >>> p = cometspy.params()
+        >>> p.set_param("writeBiomassLog", True)
+        >>> p.set_param("maxCycles", 100)
+        >>> p.set_param("BiomassLogRate",5)
+        >>> p.set_param("defaultVmax", 10.24)
+
+        """
         if name in self.all_params:
             self.all_params[name] = value
         else:
             print('Parameter ' + name + ' does not exist')
 
-    def get_param(self, name):
+    def get_param(self, name : str) -> object:
+        """
+        returns the value associated with the given parameter name
+
+        Parameters
+        ----------
+        name : str
+            the name of the parameter whose value is desired
+
+        Returns
+        -------
+        object
+            the type of object depends on the parameter requested
+
+        """
         if name in self.all_params:
             return self.all_params[name]
         else:
             print('Parameter ' + name + ' does not exist')
            
-    def __init__(self, global_params=None, package_params=None):
+    def __init__(self, global_params : str =None, 
+                 package_params : str =None):
         self.all_params = {'writeSpecificMediaLog': False,
                            'specificMediaLogRate': 1,
                            'specificMedia': 'ac_e',
@@ -245,7 +331,7 @@ class params:
                             self.all_params[k.strip()] = False
                         elif v.strip().isdigit():
                             self.all_params[k.strip()] = int(v.strip())
-                        elif isfloat(v.strip()):
+                        elif __isfloat(v.strip()):
                             self.all_params[k.strip()] = float(v.strip())
                         else:
                             self.all_params[k.strip()] = v.strip()
@@ -261,7 +347,7 @@ class params:
                             self.all_params[k.strip()] = False
                         elif v.strip().isdigit():
                             self.all_params[k.strip()] = int(v.strip())
-                        elif isfloat(v.strip()):
+                        elif __isfloat(v.strip()):
                             self.all_params[k.strip()] = float(v.strip())
                         else:
                             self.all_params[k.strip()] = v.strip()
@@ -272,8 +358,24 @@ class params:
             self.all_params['writeTotalBiomassLog'] = False
             self.all_params['writeBiomassLog'] = True
                     
-    ''' write parameters files; method probably only used by class comets'''
-    def write_params(self, out_glb, out_pkg):
+    def write_params(self, out_glb : str, out_pkg : str, to_append : str=""):
+        """
+        writes params data to the specified files
+        
+        Usually this is only used internally, though a user could use it to 
+        pre-generate params files.
+
+        Parameters
+        ----------
+        out_glb : str
+            the path and name of the 'global' parameters file to be written
+        out_pkg : str
+            the path and name of the 'package' parameters file to be written
+
+
+        """
+        out_glb = out_glb + to_append
+        out_pkg = out_pkg + to_append
 
         if os.path.isfile(out_glb):
             os.remove(out_glb)
